;;   Machine description for GNU compiler,
;;   for Freescale (Motorola) M68HC05 microcontrollers (256 Byte User Stack)
;;   Copyright (C) 2005-2006 Martin Kgler
;;   Contributed by Martin Kgler <mkoegler@auto.tuwien.ac.at>

;; This file is part of GCC.

;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING.  If not, write to the Free
;; Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
;; 02110-1301, USA.

(define_predicate "loram_operand"
  (match_operand 0 "memory_operand")
{
  if (GET_CODE (op) == SUBREG)
    op = SUBREG_REG (op);
  return MEM_REF_FLAGS (op) & 2;
})

(define_predicate "eeprom_operand"
  (match_operand 0 "memory_operand")
{
  if (GET_CODE (op) == SUBREG)
    op = SUBREG_REG (op);
  return MEM_REF_FLAGS (op) & 1;
})

;; operand refering to the stack

(define_predicate "memspaddr_operand"
  (match_code "mem,subreg")
{
  if (GET_CODE (op) == SUBREG)
    op = XEXP (op, 0);
  if (GET_CODE (op) != MEM)
    return 0;
  op = XEXP (op, 0);
  return spaddr_operand (op, HImode);
})


;; operand refering to a fixed memory location (eg a global variable)
(define_predicate "fixaddr_operand"
   (match_code "mem,subreg")
{
  if (GET_CODE (op) == MEM)
    {
      op = XEXP (op, 0);
      if (constant_addr (op))
  	return 1;
    }
  if (GET_CODE (op) == SUBREG && mode == GET_MODE (op))
    if (fixaddr_operand (XEXP (op, 0), VOIDmode))
      return 1;
  return 0;
})

;; operand in memory pointed to by a pointer
;; also an offset of 0-7 is possible for lowlevel RTL and
;; registers to be reloaded are also accepted as pointer

(define_predicate "pointerderef_operand"
   (match_code "mem,subreg")
{
  if (extra_constraint (op, 'Q'))
    return 1;
  if (GET_CODE (op) == MEM)
    {
      op = XEXP (op, 0);
      if (GET_CODE (op) == REG && GET_MODE (op) == HImode)
	return 1;
      if (GET_CODE (op) == ADDRESS && REG_P (XEXP (op, 0)))
	return 1;
    }
  if (GET_CODE (op) == SUBREG
      && pointerderef_operand (XEXP (op, 0), GET_MODE (XEXP (op, 0))))
    return 1;
  return 0;
})

;; immediate operand 
;; accepts also SUBREG of CONSTANT_P, which are used to represent parts of constant

(define_predicate "imm_operand"
  (ior (match_operand 0 "immediate_operand")
       (and (match_code "subreg")
            (match_test "CONSTANT_P (XEXP (op, 0)) && GET_MODE (XEXP (op, 0)) != VOIDmode"))))

;; memory pointer to be a ADDRESS construct

(define_predicate "memaddr_operand"
  (match_code "mem")
{
  op = XEXP (op, 0);
  if (GET_CODE (op) == ADDRESS)
    return 1;
  return 0;
})

;; Recognize operand OP of mode MODE used in call instructions

(define_predicate "call_insn_operand"
  (match_code "mem")
{
  rtx inside = XEXP (op, 0);
  if (register_operand (inside, HImode))
   return 1;
  if (CONSTANT_ADDRESS_P (inside))
   return 1;
  return 0;
})

;; a register

(define_predicate "regsp_operand"
  (match_operand 0 "register_operand"))

;; register or fixed memory location

(define_predicate "regmem_operand"
  (ior (match_operand 0 "regsp_operand")
       (match_operand 0 "fixaddr_operand")))


;; register or fixed memory location or immediate 

(define_predicate "regmemimm_operand"
  (ior (ior (match_operand 0 "regsp_operand")
            (match_operand 0 "imm_operand"))
       (match_operand 0 "fixaddr_operand")))

;; register or fixed memory location or memory pointer to by a base register

(define_predicate "regmempt_operand"
  (ior (ior (match_operand 0 "regmem_operand")
            (match_operand 0 "pointerderef_operand"))
       (match_operand 0 "memspaddr_operand")))


;; register or fixed memory location or stack or immediate

(define_predicate "regmemimmst_operand"
  (ior (match_operand 0 "regmemimm_operand")
       (match_operand 0 "memspaddr_operand")))



;; register or fixed memory location or memory pointer to by a base register or immediate

(define_predicate "regmemimmpt_operand"
  (ior (ior (match_operand 0 "regmemimm_operand")
            (match_operand 0 "pointerderef_operand"))
       (match_operand 0 "memspaddr_operand")))

(define_predicate "regmemptne_operand"
  (and (match_operand 0 "regmempt_operand")
	(not (match_operand 0 "eeprom_operand"))))

(define_predicate "regmemne_operand"
  (and (match_operand 0 "regmem_operand")
	(not (match_operand 0 "eeprom_operand"))))
